@@ -1,4 +1,4 @@
-__version__ = "0.17.1"
+__version__ = "0.27.1"
 
 from .accelerator import Accelerator
 from .big_modeling import (
@@ -11,9 +11,11 @@ from .big_modeling import (
     load_checkpoint_and_dispatch,
 )
 from .data_loader import skip_first_batches
+from .inference import prepare_pippy
 from .launchers import debug_launcher, notebook_launcher
 from .state import PartialState
 from .utils import (
+    AutocastKwargs,
     DeepSpeedPlugin,
     DistributedDataParallelKwargs,
     DistributedType,
