@@ -16,22 +16,26 @@ import json
 import os
 import tempfile
 import unittest
+from collections import OrderedDict
 
 import torch
 import torch.nn as nn
+from safetensors.torch import save_file
 
-from accelerate.test_utils import require_cuda, require_multi_gpu, require_safetensors
-from accelerate.test_utils.testing import require_torch_min_version
+from accelerate import init_empty_weights
+from accelerate.test_utils import require_cuda, require_huggingface_suite, require_multi_gpu
 from accelerate.utils.modeling import (
     check_device_map,
     clean_device_map,
     compute_module_sizes,
+    convert_file_size_to_int,
     find_tied_parameters,
     get_balanced_memory,
     infer_auto_device_map,
     load_checkpoint_in_model,
     load_state_dict,
     named_module_tensors,
+    retie_parameters,
     set_module_tensor_to_device,
 )
 
@@ -47,7 +51,37 @@ class ModelForTest(nn.Module):
         return self.linear2(self.batchnorm(self.linear1(x)))
 
 
-@require_torch_min_version(version="1.9.0")
+class LinearWithNonPersistentBuffers(nn.Module):
+    def __init__(self, in_features: int, out_features: int, bias: bool = True, device=None, dtype=None) -> None:
+        factory_kwargs = {"device": device, "dtype": dtype}
+        super().__init__()
+        self.in_features = in_features
+        self.out_features = out_features
+        self.register_buffer("weight", torch.empty((out_features, in_features), **factory_kwargs))
+        if bias:
+            self.register_buffer("bias", torch.empty(out_features, **factory_kwargs), persistent=False)
+        else:
+            self.register_buffer("bias", None)
+
+    def forward(self, input: torch.Tensor) -> torch.Tensor:
+        return torch.nn.functional.linear(input, self.weight, self.bias)
+
+
+class ModelSeveralDtypes(nn.Module):
+    def __init__(self):
+        super().__init__()
+        self.register_buffer("int_param", torch.randint(high=10, size=(15, 30)))
+        self.register_parameter("float_param", torch.nn.Parameter(torch.rand(10, 5)))
+
+    def forward(self, x):
+        return x + 2
+
+
+def sequential_model(num_layers):
+    layers = OrderedDict([(f"linear{i}", nn.Linear(1000, 1000)) for i in range(1, num_layers + 1)])
+    return nn.Sequential(layers)
+
+
 class ModelingUtilsTester(unittest.TestCase):
     def check_set_module_tensor_for_device(self, model, device1, device2):
         self.assertEqual(model.linear1.weight.device, torch.device(device1))
@@ -132,6 +166,16 @@ class ModelingUtilsTester(unittest.TestCase):
         set_module_tensor_to_device(model, "linear1.weight", "cpu", value=model.linear1.weight, dtype=torch.float16)
         self.assertEqual(model.linear1.weight.dtype, torch.float16)
 
+    def test_set_module_tensor_checks_shape(self):
+        model = ModelForTest()
+        tensor = torch.zeros((2, 2))
+        with self.assertRaises(ValueError) as cm:
+            set_module_tensor_to_device(model, "linear1.weight", "cpu", value=tensor)
+        self.assertEqual(
+            str(cm.exception),
+            'Trying to set a tensor of shape torch.Size([2, 2]) in "weight" (which has shape torch.Size([4, 3])), this look incorrect.',
+        )
+
     def test_named_tensors(self):
         model = nn.BatchNorm1d(4)
         named_tensors = named_module_tensors(model)
@@ -169,11 +213,61 @@ class ModelingUtilsTester(unittest.TestCase):
             ["linear1.weight", "linear1.bias", "batchnorm.weight", "batchnorm.bias", "linear2.weight", "linear2.bias"],
         )
 
+        model = LinearWithNonPersistentBuffers(10, 10)
+
+        named_tensors = named_module_tensors(model, include_buffers=True, remove_non_persistent=False)
+        self.assertListEqual([name for name, _ in named_tensors], ["weight", "bias"])
+
+        named_tensors = named_module_tensors(model, include_buffers=True, remove_non_persistent=True)
+        self.assertListEqual([name for name, _ in named_tensors], ["weight"])
+
     def test_find_tied_parameters(self):
-        model = ModelForTest()
-        self.assertDictEqual(find_tied_parameters(model), {})
+        model = sequential_model(4)
+        self.assertListEqual(find_tied_parameters(model), [])
+
         model.linear2.weight = model.linear1.weight
-        self.assertDictEqual(find_tied_parameters(model), {"linear1.weight": "linear2.weight"})
+        self.assertListEqual(find_tied_parameters(model), [["linear1.weight", "linear2.weight"]])
+
+        model.linear4.weight = model.linear1.weight
+        self.assertListEqual(find_tied_parameters(model), [["linear1.weight", "linear2.weight", "linear4.weight"]])
+
+        model = sequential_model(5)
+        model.linear1.weight = model.linear4.weight
+        model.linear2.weight = model.linear3.weight
+        model.linear5.weight = model.linear2.weight
+        tied_params = sorted(find_tied_parameters(model), key=lambda x: len(x))
+        self.assertListEqual(
+            tied_params, [["linear1.weight", "linear4.weight"], ["linear2.weight", "linear3.weight", "linear5.weight"]]
+        )
+
+        model = nn.Sequential(OrderedDict([("block1", sequential_model(4)), ("block2", sequential_model(4))]))
+        model.block1.linear1.weight = model.block2.linear1.weight
+        self.assertListEqual(find_tied_parameters(model), [["block1.linear1.weight", "block2.linear1.weight"]])
+
+    def test_retie_parameters(self):
+        model = sequential_model(2)
+        retie_parameters(model, [["linear1.weight", "linear2.weight"]])
+        self.assertIs(model.linear1.weight, model.linear2.weight)
+
+        model = sequential_model(3)
+        retie_parameters(model, [["linear1.weight", "linear2.weight", "linear3.weight"]])
+
+        self.assertIs(model.linear1.weight, model.linear2.weight)
+        self.assertIs(model.linear1.weight, model.linear3.weight)
+
+        model = sequential_model(5)
+        retie_parameters(
+            model, [["linear1.weight", "linear4.weight"], ["linear2.weight", "linear3.weight", "linear5.weight"]]
+        )
+
+        self.assertIs(model.linear1.weight, model.linear4.weight)
+        self.assertIs(model.linear2.weight, model.linear3.weight)
+        self.assertIs(model.linear2.weight, model.linear5.weight)
+
+        model = nn.Sequential(OrderedDict([("block1", sequential_model(4)), ("block2", sequential_model(4))]))
+        retie_parameters(model, [["block1.linear1.weight", "block2.linear1.weight"]])
+
+        self.assertIs(model.block1.linear1.weight, model.block2.linear1.weight)
 
     def test_compute_module_sizes(self):
         model = ModelForTest()
@@ -341,6 +435,19 @@ class ModelingUtilsTester(unittest.TestCase):
         self.assertEqual(model.batchnorm.weight.device, torch.device("cpu"))
         self.assertEqual(model.linear2.weight.device, torch.device(1))
 
+    def test_load_checkpoint_in_model_dtype(self):
+        with tempfile.NamedTemporaryFile(suffix=".pt") as tmpfile:
+            model = ModelSeveralDtypes()
+            torch.save(model.state_dict(), tmpfile.name)
+
+            new_model = ModelSeveralDtypes()
+            load_checkpoint_in_model(
+                new_model, tmpfile.name, offload_state_dict=True, dtype=torch.float16, device_map={"": "cpu"}
+            )
+
+            self.assertEqual(new_model.int_param.dtype, torch.int64)
+            self.assertEqual(new_model.float_param.dtype, torch.float16)
+
     def test_clean_device_map(self):
         # Regroup everything if all is on the same device
         self.assertDictEqual(clean_device_map({"a": 0, "b": 0, "c": 0}), {"": 0})
@@ -384,15 +491,114 @@ class ModelingUtilsTester(unittest.TestCase):
         )
         self.assertDictEqual(device_map, {"0": 0, "1": 1, "2": 1})
 
-        # Now if we have weights tied inside submodules, tied weights are on the same device.
-        model = nn.Sequential(ModelForTest(), ModelForTest(), ModelForTest())
-        layer0 = getattr(model, "0")
-        layer2 = getattr(model, "2")
-        layer0.linear2.weight = layer2.linear2.weight
+    def test_infer_auto_device_map_with_tied_weights(self):
+        model = nn.Sequential(
+            OrderedDict([("layer1", ModelForTest()), ("layer2", ModelForTest()), ("layer3", ModelForTest())])
+        )
+        model.layer3.linear2.weight = model.layer1.linear2.weight
         device_map = infer_auto_device_map(model, max_memory={0: 400, 1: 500})
-        expected = {"0": 0, "2.linear2": 0, "1": 1, "2.linear1": 1, "2.batchnorm": 1}
+        expected = {"layer1": 0, "layer3.linear2": 0, "layer2": 1, "layer3.linear1": 1, "layer3.batchnorm": 1}
         self.assertDictEqual(device_map, expected)
 
+        # With three weights tied together
+        model.layer2.linear2.weight = model.layer1.linear2.weight
+        device_map = infer_auto_device_map(model, max_memory={0: 400, 1: 500})
+        expected = {
+            "layer1": 0,
+            "layer2.linear2": 0,
+            "layer3.linear2": 0,
+            "layer2.linear1": 1,
+            "layer2.batchnorm": 1,
+            "layer3.linear1": 1,
+            "layer3.batchnorm": 1,
+        }
+        self.assertDictEqual(device_map, expected)
+
+        # With two groups of weights tied together
+        model.layer2.linear1.weight = model.layer1.linear1.weight
+        device_map = infer_auto_device_map(model, max_memory={0: 400, 1: 500})
+        expected = {
+            "layer1": 0,
+            "layer2.linear1": 0,
+            "layer2.linear2": 0,
+            "layer3.linear2": 0,
+            "layer2.batchnorm": 1,
+            "layer3.linear1": 1,
+            "layer3.batchnorm": 1,
+        }
+        self.assertDictEqual(device_map, expected)
+
+        # With weights ties in the same module
+        model = nn.Sequential(
+            OrderedDict(
+                [
+                    ("linear1", nn.Linear(4, 4)),
+                    ("linear2", nn.Linear(6, 6)),
+                    ("linear3", nn.Linear(4, 4)),
+                    ("linear4", nn.Linear(6, 6)),
+                ]
+            )
+        )
+        model.linear3.weight = model.linear1.weight
+        model.linear3.bias = model.linear1.bias
+        device_map = infer_auto_device_map(model, max_memory={0: 250, 1: 400})
+        expected = {"linear1": 0, "linear2": 1, "linear3": 0, "linear4": 1}
+        self.assertDictEqual(device_map, expected)
+
+        # With tied weights sharing a same prefix name (`compute.weight` vs `compute.weight_submodule.parameter`)
+        class SubModule(torch.nn.Module):
+            def __init__(self, ref_to_parameter):
+                super().__init__()
+                self.parameter = ref_to_parameter
+
+            def forward(self, x):
+                return self.x + torch.max(self.parameter)
+
+        class LinearModuleAndSubModule(torch.nn.Linear):
+            def __init__(self, in_features, out_features):
+                super().__init__(in_features, out_features)
+                self.weight_submodule = SubModule(self.weight)
+
+            def forward(self, x):
+                return torch.nn.functional.linear(self.weight_submodule(x), self.weight)
+
+        class Model(torch.nn.Module):
+            def __init__(self):
+                super().__init__()
+                self.compute = LinearModuleAndSubModule(3, 8)
+
+            def forward(self, x):
+                return self.compute(x)
+
+        model = Model()
+
+        device_memory = {0: 4, "cpu": 96000}  # Low memory device, just to force splitting and trigger the error
+        infer_auto_device_map(model, device_memory)
+
+    @require_huggingface_suite
+    def test_infer_auto_device_map_on_t0pp(self):
+        from transformers import AutoConfig, AutoModelForSeq2SeqLM
+
+        config = AutoConfig.from_pretrained("bigscience/T0pp")
+        with init_empty_weights():
+            model = AutoModelForSeq2SeqLM.from_config(config)
+        model.tie_weights()
+
+        special_dtypes = {n: torch.float32 for n, _ in model.named_parameters() if "wo" in n}
+        max_memory = {0: 10**10, 1: 10**10, "cpu": 10**10}
+        device_map = infer_auto_device_map(
+            model,
+            no_split_module_classes=["T5Block"],
+            dtype=torch.float16,
+            max_memory=max_memory,
+            special_dtypes=special_dtypes,
+        )
+
+        # The 3 tied weights should all be on device 0
+        self.assertEqual(device_map["shared"], 0)
+        self.assertEqual(device_map["encoder.embed_tokens"], 0)
+        self.assertEqual(device_map["decoder.embed_tokens"], 0)
+
     @require_cuda
     def test_get_balanced_memory(self):
         model = ModelForTest()
@@ -400,6 +606,10 @@ class ModelingUtilsTester(unittest.TestCase):
         max_memory = get_balanced_memory(model, max_memory={0: 200, 1: 200})
         self.assertDictEqual({0: 200, 1: 200}, max_memory)
 
+        # We should be able to set models on a non-contiguous sub-set of
+        max_memory = get_balanced_memory(model, max_memory={0: 200, 2: 200})
+        self.assertDictEqual({0: 200, 2: 200}, max_memory)
+
         max_memory = get_balanced_memory(model, max_memory={0: 300, 1: 300})
         self.assertDictEqual({0: 215, 1: 300}, max_memory)
 
@@ -415,11 +625,12 @@ class ModelingUtilsTester(unittest.TestCase):
         max_memory = get_balanced_memory(model, max_memory={0: 0, 1: 300, 2: 300})
         self.assertDictEqual({0: 0, 1: 215, 2: 300}, max_memory)
 
-    @require_cuda
-    @require_safetensors
-    def test_load_state_dict(self):
-        from safetensors.torch import save_file
+        # If we set a device to 0, it's not counted.
+        max_memory = get_balanced_memory(model, max_memory={0: 0, "cpu": 100})
+        self.assertDictEqual({0: 0, "cpu": 100}, max_memory)
 
+    @require_cuda
+    def test_load_state_dict(self):
         state_dict = {k: torch.randn(4, 5) for k in ["a", "b", "c"]}
         device_maps = [{"a": "cpu", "b": 0, "c": "disk"}, {"a": 0, "b": 0, "c": "disk"}, {"a": 0, "b": 0, "c": 0}]
 
@@ -433,3 +644,31 @@ class ModelingUtilsTester(unittest.TestCase):
             for param, device in device_map.items():
                 device = device if device != "disk" else "cpu"
                 self.assertEqual(loaded_state_dict[param].device, torch.device(device))
+
+    def test_convert_file_size(self):
+        result = convert_file_size_to_int("100MB")
+        self.assertEqual(result, 100 * (10**6))
+
+        result = convert_file_size_to_int("2GiB")
+        self.assertEqual(result, 2 * (2**30))
+
+        result = convert_file_size_to_int("512KiB")
+        self.assertEqual(result, 512 * (2**10))
+
+        result = convert_file_size_to_int("1.5GB")
+        self.assertEqual(result, 1.5 * (10**9))
+
+        result = convert_file_size_to_int("100KB")
+        self.assertEqual(result, 100 * (10**3))
+
+        result = convert_file_size_to_int(500)
+        self.assertEqual(result, 500)
+
+        with self.assertRaises(ValueError):
+            convert_file_size_to_int("5MBB")
+
+        with self.assertRaises(ValueError):
+            convert_file_size_to_int("5k0MB")
+
+        with self.assertRaises(ValueError):
+            convert_file_size_to_int("-1GB")
